#include "squidwall.h"

static int read_dir2array(array *root, const char *path, time_t *mtime, int type);

static array array_hosts, array_users,
 array_pfilter, array_bfilter, array_cfilter, array_ufilter, array_wfilter,
 array_bcfilter, array_brfilter, array_crfilter, array_urfilter, array_wrfilter;

static time_t mtime_hosts, mtime_users,
 mtime_pfilter, mtime_bfilter, mtime_cfilter, mtime_ufilter, mtime_wfilter,
 mtime_bcfilter, mtime_brfilter, mtime_crfilter, mtime_urfilter, mtime_wrfilter;

/* todolist of headers / mimetypes which should be scanned */
array array_cfilter_todo;
array array_pfilter_todo;

struct squidwall_acl *acl;
struct squidwall_phrase *phrase;
struct squidwall_string *string;
struct squidwall_regex *regex;
struct named_regex *nr;

static void read_acls_regex_options(struct named_regex *x, const char *path, char *file)
{
	stralloc sa;
	buffer b;
	int plus=0;

	stralloc_init(&sa);

	if (!stralloc_copys(&sa, path)) die_nomem();
	if (!stralloc_cats(&sa, "/")) die_nomem();
	if (!stralloc_cats(&sa, file)) die_nomem();
	sa.s[sa.len]=0;

	if (buffer_mmapread(&b, sa.s) == -1) {
		stralloc_free(&sa); return;
	}

	while (buffer_getnewline_sa(&b, &sa) == 1) {

		/* skip empty lines */
		if (sa.len == 1) continue;

		/* skip comments */
		if (sa.s[0] == '#') continue;
		sa.s[sa.len] = 0; /* ensure termination */

		/* remove \r\n and \n */
		if (sa.s[sa.len-2] == '\r') sa.s[sa.len-2]=0;
		if (sa.s[sa.len-1] == '\n') sa.s[sa.len-1]=0;

		switch (sa.s[0]) {
		case '+':
			plus = 1;
			break;
		case '-':
			plus = 0;
			break;
		default:
			continue;
		}

		if (str_start(sa.s+1, "REG_EXTENDED") && plus) {
			set(x->cflags, REG_EXTENDED);
		} else {
			unset(x->cflags, REG_EXTENDED);
		}

		if (str_start(sa.s+1, "REG_ICASE") && plus) {
			set(x->cflags, REG_EXTENDED);
		} else {
			unset(x->cflags, REG_EXTENDED);
		}

		if (str_start(sa.s+1, "REG_NOSUB") && plus) {
			set(x->cflags, REG_NOSUB);
		} else {
			unset(x->cflags, REG_NOSUB);
		}

		if (str_start(sa.s+1, "REG_NEWLINE") && plus) {
			set(x->cflags, REG_NEWLINE);
		} else {
			unset(x->cflags, REG_NEWLINE);
		}

		/* eflags for regexec() */
		if (str_start(sa.s+1, "REG_NOTBOL") && plus) {
			set(x->eflags, REG_NOTBOL);
		} else {
			unset(x->eflags, REG_NOTBOL);
		}

		if (str_start(sa.s+1, "REG_NOTEOL") && plus) {
			set(x->eflags, REG_NOTEOL);
		} else {
			unset(x->eflags, REG_NOTEOL);
		}

	}
	buffer_close(&b);
	stralloc_free(&sa);

	return;
}

static void read_acls_regex(array *root, const char *path, char *file)
{
	stralloc sa;
	buffer b;

	stralloc_init(&sa);

	if (!stralloc_copys(&sa, path)) die_nomem();
	if (!stralloc_cats(&sa, "/")) die_nomem();
	if (!stralloc_cats(&sa, file)) die_nomem();
	sa.s[sa.len]=0;

	if (buffer_mmapread(&b, sa.s) == -1) {
		stralloc_free(&sa); return;
	}

	regex=array_allocate(root, sizeof(struct squidwall_regex),
			array_length(root, sizeof(struct squidwall_regex)));

	if (!regex) die_nomem();
	byte_zero(regex, sizeof(struct squidwall_regex));
	INIT_LIST_HEAD(&regex->list);

	/* save the filename of the filter */
	stralloc_init(&regex->filename);
	if (!stralloc_copys(&regex->filename, file)) die_nomem();
	regex->filename.s[regex->filename.len]=0;

	while (buffer_getnewline_sa(&b, &sa) == 1) {

		/* skip empty lines */
		if (sa.len == 1) continue;

		/* skip comments */
		if (sa.s[0] == '#') continue;
		sa.s[sa.len] = 0; /* ensure termination */

		/* remove \r\n and \n */
		if (sa.s[sa.len-2] == '\r') sa.s[sa.len-2]=0;
		if (sa.s[sa.len-1] == '\n') sa.s[sa.len-1]=0;

		/* save the precompiled regex */
		nr=array_allocate(&regex->regex, sizeof(struct named_regex),
				array_length(&regex->regex, sizeof(struct named_regex)));

		if (!nr) die_nomem();
		byte_zero(nr, sizeof(struct named_regex));
		read_acls_regex_options(nr, path, "@");
		if (regcomp(&nr->r, sa.s, nr->cflags) != 0) {
			log_headline("error");
			log_s("regcomp() failed @ this line: ["); log_s(sa.s); log_s("]\n");
			continue; /* read next line */
		}

		/* save the string */
		stralloc_init(&nr->n);
		if (!stralloc_copys(&nr->n, sa.s)) die_nomem();
	}
	buffer_close(&b);
	stralloc_free(&sa);

	return;
}

static void read_acls_string(array *root, const char *path, char *file)
{
	stralloc sa;
	buffer b;

	stralloc_init(&sa);

	if (!stralloc_copys(&sa, path)) die_nomem();
	if (!stralloc_cats(&sa, "/")) die_nomem();
	if (!stralloc_cats(&sa, file)) die_nomem();
	sa.s[sa.len]=0;

	if (buffer_mmapread(&b, sa.s) == -1) {
		stralloc_free(&sa); return;
	}

	/* add string to filter */
	string=array_allocate(root, sizeof(struct squidwall_string),
			array_length(root, sizeof(struct squidwall_string)));

	if (!string) die_nomem();
	byte_zero(string, sizeof(struct squidwall_string));
	INIT_LIST_HEAD(&string->list);

	/* save the filename of the filter */
	stralloc_init(&string->filename);
	if (!stralloc_copys(&string->filename, file)) die_nomem();
	string->filename.s[string->filename.len]=0;

	/* save all the strings to search for in the stralloc 'strings' */
	stralloc_init(&string->strings);
	while (buffer_getnewline_sa(&b, &sa) == 1) {

		/* skip empty lines */
		if (sa.len == 1) continue;

		/* skip comments */
		if (sa.s[0] == '#') continue;
		sa.s[sa.len] = 0; /* ensure termination */

		/* remove \r\n and \n */
		if (sa.s[sa.len-2] == '\r') sa.s[sa.len-2]=0;
		if (sa.s[sa.len-1] == '\n') sa.s[sa.len-1]=0;

		if (!stralloc_cats(&string->strings, sa.s)) die_nomem();
		if (!stralloc_0(&string->strings)) die_nomem();
	}
	buffer_close(&b);
	stralloc_free(&sa);

	return;
}

static void read_acls_phrase(array *root, const char *path, char *file)
{
	stralloc sa;
	buffer b;
	int i, state, count=0;
	char *str, *num;
	unsigned int *ui;

	stralloc_init(&sa);

	if (!stralloc_copys(&sa, path)) die_nomem();
	if (!stralloc_cats(&sa, "/")) die_nomem();
	if (!stralloc_cats(&sa, file)) die_nomem();
	sa.s[sa.len]=0;

	if (buffer_mmapread(&b, sa.s) == -1) {
		stralloc_free(&sa); return;
	}

	/* add string to filter */
	phrase=array_allocate(root, sizeof(struct squidwall_phrase),
			array_length(root, sizeof(struct squidwall_phrase)));

	if (!phrase) die_nomem();
	byte_zero(phrase, sizeof(struct squidwall_phrase));
	INIT_LIST_HEAD(&phrase->list);

	/* save the filename of the filter */
	stralloc_init(&phrase->filename);
	if (!stralloc_copys(&phrase->filename, file)) die_nomem();
	phrase->filename.s[phrase->filename.len]=0;

	/* save all the strings to search for in the stralloc 'strings' */
	stralloc_init(&phrase->strings);
	array_reset(&phrase->points);
	while (buffer_getnewline_sa(&b, &sa) == 1) {

		/* skip empty lines */
		if (sa.len == 1) continue;

		/* skip comments */
		if (sa.s[0] == '#') continue;
		sa.s[sa.len] = 0; /* ensure termination */

		/* remove \r\n and \n */
		if (sa.s[sa.len-2] == '\r') sa.s[sa.len-2]=0;
		if (sa.s[sa.len-1] == '\n') sa.s[sa.len-1]=0;

		/* parsing of "< string ><num>" */
		for (i=0, state=0, num=0, str=0; i<=sa.len; i++) {
			switch (state) {
			case 0:
			if (sa.s[i] == '<') {
				sa.s[i]=0; state=1; str=sa.s+i+1;
			}
			break;
			case 1:
			if (sa.s[i] == '>') {
				sa.s[i]=0; state=2; sa.s[i]=0;
			}
			break;
			case 2:
			if (sa.s[i] == '<') {
				sa.s[i]=0; state=3; num=sa.s+i+1;
			}
			break;
			case 3:
			if (sa.s[i] == '>') {
				sa.s[i]=0; state=4; sa.s[i]=0;
			}
			break;
			} /* switch() */
		}

		if (state == 4) {
			/* sth. like "<lala><33>" --> was ok */
			ui=array_allocate(&phrase->points, sizeof(unsigned int), count);
			if (!ui) die_nomem();
			if (!scan_uint(num, ui)) *ui = PFILTER_DEFAULT;
			if (!stralloc_cats(&phrase->strings, str)) die_nomem();
			if (!stralloc_0(&phrase->strings)) die_nomem();
			count++;
		}
	}
	buffer_close(&b);
	stralloc_free(&sa);

	return;
}

/**
 * read_dir2array - read a directory
 * @path:	the path, which should be read
 * @root:	pointer of the array
 * @mtime:	the last modification time
 * @type:	the squidwall type: acl, string, regex
 *
 * 1. check if '!' was modified
 * 2. if modified, scan the whole directory to an array of type X
 * 3. if the acl is a filtering one, scan the filteroptions
 */

static int read_dir2array(array *root, const char *path, time_t *mtime, int type)
{
	struct dirent *d;
	struct stat s;
	stralloc sa;
	DIR *D;
	int i, j;

	stralloc_init(&sa);
	if (!stralloc_copys(&sa, path)) die_nomem();
	if (!stralloc_cats(&sa, "/!")) die_nomem();
	if (!stralloc_0(&sa)) die_nomem();

	/* if stat fails, we return ok -> no acl / filter directory */
	if (stat(sa.s, &s) == -1) { stralloc_free(&sa); return -1; }
	stralloc_free(&sa);

	/* was not modified... */
	if (*mtime == s.st_mtime) { return 0; }
	*mtime = s.st_mtime;

	/* free the old array */
	switch (type) {
	case TYPE_ACL:
		for (i=0; i < array_length(root, sizeof(struct squidwall_acl)); i++) {
			acl = array_get(root, sizeof(struct squidwall_acl), i);
			stralloc_free(&acl->filename);
			stralloc_free(&acl->bfilter);
			stralloc_free(&acl->bcfilter);
			stralloc_free(&acl->brfilter);
			stralloc_free(&acl->cfilter);
			stralloc_free(&acl->crfilter);
			stralloc_free(&acl->pfilter);
			stralloc_free(&acl->ufilter);
			stralloc_free(&acl->urfilter);
			stralloc_free(&acl->wfilter);
			stralloc_free(&acl->wrfilter);
		}
		array_reset(root);
		break;
	case TYPE_STRING:
		for (i=0; i < array_length(root, sizeof(struct squidwall_string)); i++) {
			string = array_get(root, sizeof(struct squidwall_string), i);
			stralloc_free(&string->filename);
			stralloc_free(&string->strings);
		}
		array_reset(root);
		break;
	case TYPE_PHRASE:
		for (i=0; i < array_length(root, sizeof(struct squidwall_phrase)); i++) {
			phrase = array_get(root, sizeof(struct squidwall_phrase), i);
			stralloc_free(&phrase->filename);
			stralloc_free(&phrase->strings);
		}
		array_reset(root);
		break;
	case TYPE_REGEX:
		for (i=0; i < array_length(root, sizeof(struct squidwall_regex)); i++) {
			regex = array_get(root, sizeof(struct squidwall_regex), i);
			stralloc_free(&regex->filename);
			for (j=0; j < array_length(&regex->regex, sizeof(struct named_regex)); j++) {
				nr = array_get(&regex->regex, sizeof(struct named_regex), j);
				regfree(&nr->r);
				stralloc_free(&nr->n);
			}
			array_reset(&regex->regex);
		}
		array_reset(root);
		break;
	}

	/* open directory or return */
	D=opendir(path); if (!D) return -1;
	i = 0;

	/* read it */
	while ((d=readdir(D))) {


		/* skip '.' and '..' */
		if (d->d_name[0]=='.' && d->d_name[1]==0) continue;
		if (d->d_name[0]=='.' && d->d_name[1]=='.' && d->d_name[2]==0)
			continue;

		/* skip '@' (contains regex options) */
		if (d->d_name[0]=='@' && d->d_name[1]==0) continue;

		/* skip '!' */
		if (d->d_name[0]=='!' && d->d_name[1]==0) continue;

		/* skip this pecial file */
		if (str_start(d->d_name, "README")) continue;

		switch (type) {
		case TYPE_ACL:

			/* filename is to short */
			if (d->d_reclen <= 1) continue;

			/* filename begins with a false character */
			if (d->d_name[0]!='+' && d->d_name[0]!='-' &&
				d->d_name[0]!='?') continue;

			acl=array_allocate(root, sizeof(struct squidwall_acl), i);
			if (!acl) die_nomem();
			byte_zero(acl, sizeof(struct squidwall_acl));
			if(!stralloc_copys(&acl->filename, d->d_name+1)) die_nomem();
			acl->filename.s[acl->filename.len]=0;

			INIT_LIST_HEAD(&acl->list);

			stralloc_init(&acl->bfilter);
			stralloc_init(&acl->bcfilter);
			stralloc_init(&acl->brfilter);
			stralloc_init(&acl->cfilter);
			stralloc_init(&acl->crfilter);
			stralloc_init(&acl->pfilter);
			stralloc_init(&acl->ufilter);
			stralloc_init(&acl->urfilter);
			stralloc_free(&acl->wfilter);
			stralloc_free(&acl->wrfilter);

			/* get names */
			switch (d->d_name[0]) {
			case '+':
				/* ALLOW all */
				acl->flags=ACL_ALLOW;
				break;
			case '-':
				/* DENY all */
				acl->flags=ACL_DENY;
				break;
			case '?':
				/* filtered / scanned */
				acl->flags=ACL_FILTER;
				break;
			}
			break;


		case TYPE_REGEX:
			read_acls_regex(root, path, d->d_name);
			break;

		case TYPE_STRING:
			read_acls_string(root, path, d->d_name);
			break;

		case TYPE_PHRASE:
			read_acls_phrase(root, path, d->d_name);
			break;
		}
		i++;
	}
	closedir(D);

	return i;
}


/**
 * read_acls - read the acl directories
 *
 * 1. read all acl directories into the arrays and setup these global lists:
 *    - cfilter => all content filters
 *    - pfilter => all phrase filters
 *    - ufilter => all url filters
 *    - crfilter => all content-regex filters
 *    - urfilter => all url regex filters
 *    - hosts => all hosts which should be filtered
 *    - users => all users which should be filtered
 * 3. setup the filter lists for users/hosts
 */

void read_acls(void)
{
	int i, count, reload=0, logged=0;

	i=read_dir2array(&array_bfilter, bfilter_dir, &mtime_bfilter, TYPE_STRING);
	if (i) {
		INIT_LIST_HEAD(&bfilter);
		for (i=0; i < array_length(&array_bfilter, sizeof(struct squidwall_string)); i++) {
			string = array_get(&array_bfilter, sizeof(struct squidwall_string), i);
			list_add(&string->list, &bfilter);
		}
#if WANT_STARTUP_INFO
		if (i) {
			log_headline("reading"); log_s("bfilter:  "); log_ulong(i); log_s("\n");
			logged=1;
		}
#endif
		reload=1;
	}

	i=read_dir2array(&array_cfilter, cfilter_dir, &mtime_cfilter, TYPE_STRING);
	if (i) {
		INIT_LIST_HEAD(&cfilter);
		for (i=0; i < array_length(&array_cfilter, sizeof(struct squidwall_string)); i++) {
			string = array_get(&array_cfilter, sizeof(struct squidwall_string), i);
			list_add(&string->list, &cfilter);
		}
#if WANT_STARTUP_INFO
		if (i) {
			log_headline("reading"); log_s("cfilter:  "); log_ulong(i); log_s("\n");
			logged=1;
		}
#endif
		reload=1;
		read_acls_regex(&array_cfilter_todo, cfilter_dir, "!");
	}

	i=read_dir2array(&array_pfilter, pfilter_dir, &mtime_pfilter, TYPE_PHRASE);
	if (i) {
		INIT_LIST_HEAD(&pfilter);
		for (i=0; i < array_length(&array_pfilter, sizeof(struct squidwall_phrase)); i++) {
			phrase = array_get(&array_pfilter, sizeof(struct squidwall_phrase), i);
			list_add(&phrase->list, &pfilter);
		}
#if WANT_STARTUP_INFO
		if (i) {
			log_headline("reading"); log_s("pfilter:  "); log_ulong(i); log_s("\n");
			logged=1;
		}
#endif
		reload=1;
		read_acls_regex(&array_pfilter_todo, pfilter_dir, "!");
	}
	
	i=read_dir2array(&array_ufilter, ufilter_dir, &mtime_ufilter, TYPE_STRING);
	if (i) {
		INIT_LIST_HEAD(&ufilter);
		for (i=0; i < array_length(&array_ufilter, sizeof(struct squidwall_string)); i++) {
			string = array_get(&array_ufilter, sizeof(struct squidwall_string), i);
			list_add(&string->list, &ufilter);
		}
#if WANT_STARTUP_INFO
		if (i) {
			log_headline("reading"); log_s("ufilter:  "); log_ulong(i); log_s("\n");
			logged=1;
		}
#endif
		reload=1;
	}

	i=read_dir2array(&array_wfilter, wfilter_dir, &mtime_wfilter, TYPE_STRING);
	if (i) {
		INIT_LIST_HEAD(&wfilter);
		for (i=0; i < array_length(&array_wfilter, sizeof(struct squidwall_string)); i++) {
			string = array_get(&array_wfilter, sizeof(struct squidwall_string), i);
			list_add(&string->list, &wfilter);
		}
#if WANT_STARTUP_INFO
		if (i) {
			log_headline("reading"); log_s("wfilter:  "); log_ulong(i); log_s("\n");
			logged=1;
		}
#endif
		reload=1;
	}

	i=read_dir2array(&array_bcfilter, bcfilter_dir, &mtime_bcfilter, TYPE_REGEX);
	if (i) {
		INIT_LIST_HEAD(&bcfilter);
		for (i=0; i < array_length(&array_bcfilter, sizeof(struct squidwall_regex)); i++) {
			regex = array_get(&array_bcfilter, sizeof(struct squidwall_regex), i);
			list_add(&regex->list, &bcfilter);
		}
#if WANT_STARTUP_INFO
		if (i) {
			log_headline("reading"); log_s("bcfilter: "); log_ulong(i); log_s("\n");
			logged=1;
		}
#endif
		reload=1;
	}

	i=read_dir2array(&array_brfilter, brfilter_dir, &mtime_brfilter, TYPE_REGEX);
	if (i) {
		INIT_LIST_HEAD(&brfilter);
		for (i=0; i < array_length(&array_brfilter, sizeof(struct squidwall_regex)); i++) {
			regex = array_get(&array_brfilter, sizeof(struct squidwall_regex), i);
			list_add(&regex->list, &brfilter);
		}
#if WANT_STARTUP_INFO
		if (i) {
			log_headline("reading"); log_s("brfilter: "); log_ulong(i); log_s("\n");
			logged=1;
		}
#endif
		reload=1;
	}

	i=read_dir2array(&array_crfilter, crfilter_dir, &mtime_crfilter, TYPE_REGEX);
	if (i) {
		INIT_LIST_HEAD(&crfilter);
		for (i=0; i < array_length(&array_crfilter, sizeof(struct squidwall_regex)); i++) {
			regex = array_get(&array_crfilter, sizeof(struct squidwall_regex), i);
			list_add(&regex->list, &crfilter);
		}
#if WANT_STARTUP_INFO
		if (i) {
			log_headline("reading"); log_s("crfilter: "); log_ulong(i); log_s("\n");
			logged=1;
		}
#endif
		reload=1;
	}

	i=read_dir2array(&array_urfilter, urfilter_dir, &mtime_urfilter, TYPE_REGEX);
	if (i) {
		INIT_LIST_HEAD(&urfilter);
		for (i=0; i < array_length(&array_urfilter, sizeof(struct squidwall_regex)); i++) {
			regex = array_get(&array_urfilter, sizeof(struct squidwall_regex), i);
			list_add(&regex->list, &urfilter);
		}
#if WANT_STARTUP_INFO
		if (i) {
			log_headline("reading"); log_s("urfilter: "); log_ulong(i); log_s("\n");
			logged=1;
		}
#endif
		reload=1;
	}

	i=read_dir2array(&array_wrfilter, wrfilter_dir, &mtime_wrfilter, TYPE_REGEX);
	if (i) {
		INIT_LIST_HEAD(&wrfilter);
		for (i=0; i < array_length(&array_wrfilter, sizeof(struct squidwall_regex)); i++) {
			regex = array_get(&array_wrfilter, sizeof(struct squidwall_regex), i);
			list_add(&regex->list, &wrfilter);
		}
#if WANT_STARTUP_INFO
		if (i) {
			log_headline("reading"); log_s("wrfilter: "); log_ulong(i); log_s("\n");
			logged=1;
		}
#endif
		reload=1;
	}

	i=read_dir2array(&array_users, user_dir, &mtime_users, TYPE_ACL);
	if (reload || i) {
		INIT_LIST_HEAD(&users);
		count=array_length(&array_users, sizeof(struct squidwall_acl));
		for (i=0; i < count; i++) {
			acl = array_get(&array_users, sizeof(struct squidwall_acl), i);
			list_add(&acl->list, &users); /* add the acl to the users list */
			
			if (!stralloc_copys(&acl->bfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->bcfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->brfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->cfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->crfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->pfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->ufilter, "")) die_nomem();
			if (!stralloc_copys(&acl->urfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->wfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->wrfilter, "")) die_nomem();

			if (acl->flags == ACL_FILTER)
				read_filters(acl, user_dir, acl->filename.s);

		}
#if WANT_STARTUP_INFO
		if (i) {
			log_headline("reading"); log_s("users:    "); log_ulong(i); log_s("\n");
			logged=1;
		}
#endif
	}

	i=read_dir2array(&array_hosts, host_dir, &mtime_hosts, TYPE_ACL);
	if (reload || i) {
		INIT_LIST_HEAD(&hosts);
		count=array_length(&array_hosts, sizeof(struct squidwall_acl));
		for (i=0; i < count; i++) {
			acl = array_get(&array_hosts, sizeof(struct squidwall_acl), i);
			list_add(&acl->list, &hosts); /* add the acl to the hosts list */

			if (!stralloc_copys(&acl->bfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->bcfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->brfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->cfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->crfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->pfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->ufilter, "")) die_nomem();
			if (!stralloc_copys(&acl->urfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->wfilter, "")) die_nomem();
			if (!stralloc_copys(&acl->wrfilter, "")) die_nomem();

			if (acl->flags == ACL_FILTER)
				read_filters(acl, host_dir, acl->filename.s);

		}
#if WANT_STARTUP_INFO
		if (i) {
			log_headline("reading"); log_s("hosts:    "); log_ulong(i); log_s("\n");
			logged=1;
		}
#endif
	}

	if (logged) {
		log_headline("reading");
		for (i=0; i < 38; i++) log_s("-");
		log_s("\n");
	}

	return;
}
#
