#include "squidwall.h"

/**
 * read_filters - read filteroptions
 * @acl:  the access control list
 * @path: the acl path ($SW/hosts or $SW/users)
 * @file: filename for reading
 *
 * ---------------------------------------------------
 * banner filtering
 * 
 * +b,Filter1,Filter2,Filter3
 * -b disable banner filtering
 *
 * banner filtering, using regex.7 expressions
 * +B,Filter1,Filter2,Filter3
 * -B disable regex banner filtering
 *
 * banner filtering, using regex.7 @ http header
 * +a,Filter1,Filter2,Filter3
 * -a disable regex banner filtering
 *
 * ---------------------------------------------------
 * url blacklist
 * 
 * +u,Filter1,Filter2,Filter3
 * -u disable url filtering
 *
 * url filtering, using regex.7 expressions
 * +U,Filter1,Filter2,Filter3
 * -U disable regex url filtering
 *
 * ---------------------------------------------------
 * url whitelist
 * 
 * +w,Filter1,Filter2,Filter3
 * -w disable url filtering
 *
 * url whitelist, using regex.7 expressions
 * +W,Filter1,Filter2,Filter3
 * -W disable regex url filtering
 *
 * ---------------------------------------------------
 * content filtering
 * 
 * +c,Filter1,Filter2,Filter3
 * -c disable content filtering
 *
 * content filtering, using regex.7 expressions
 * +C,Filter1,Filter2,Filter3
 * -C disable regex content filtering
 *
 * phrase filtering (strings with hitpoints)
 * +p,Filter1,Filter2,Filter3
 * -p disable content filtering
 *
 * antivirus scanning
 * +V
 * -V disable antivirus scanning
 */

static int state_bfilter, state_bcfilter, state_brfilter,
	   state_cfilter, state_pfilter, state_crfilter,
	   state_ufilter, state_urfilter,
	   state_wfilter, state_wrfilter;

#define STATE_NONE 0 /* no file in filter-definition */
#define STATE_SOME 1 /* some files */
#define STATE_ALL  2 /* all files need to be read! */

void read_filters_addfile(stralloc *sa, int *state, char *start)
{
	char *x=start;
	int i, l;

	if (!x) return;
	if (!x[1]) return;
	if (x[0] != ',') return;

	i = 0; l = str_len(start);
	while (*++x) {

		/* read all available filters and return */
		if (*x == '*') {
			*state = STATE_ALL;
			return;
		}

		/* ',' -> '\0' */
		if (*x == ',') *x = 0;
	}

	/* read and compile each given filter */
	for (x=start+1; i < l; i+=str_len(x+i)+1) {
		if(!stralloc_cats(sa, x+i)) die_nomem();
		if(!stralloc_0(sa)) die_nomem();
		*state = STATE_SOME;
	}

	return;
}

void read_filters(struct squidwall_acl *acl, const char *path, const char *file)
{
	struct squidwall_phrase *phrase;
	struct squidwall_string *string;
	struct squidwall_regex *regex;
	struct list_head *l;
	int todo_done;
	stralloc sa;
	buffer b;

	stralloc_init(&sa);

	if (!stralloc_copys(&sa, path)) die_nomem();
	if (!stralloc_cats(&sa, "/?")) die_nomem();
	if (!stralloc_cats(&sa, file)) die_nomem();
	sa.s[sa.len]=0;

	/* read /etc/squidwall/hosts/+192.168.100.5 e.g. */
	if (buffer_mmapread(&b, sa.s) == -1) {
		/* if the host/user failed... read the '!' */
		if (!stralloc_copys(&sa, path)) die_nomem();
		if (!stralloc_cats(&sa, "/!")) die_nomem();
		sa.s[sa.len]=0;
		if (buffer_mmapread(&b, sa.s) == -1) return;
	}

	/* nothing was read currently */
	todo_done=TODO_NONE;
	acl->todo=ACL_DEFAULT;

	/* what is wanted */
	for (;;) {
		/* 0=EOF -1=error 1=ok */
		if (buffer_getnewline_sa(&b, &sa) != 1) break;
		sa.s[sa.len-1]=0; /* remove newline */

		if (sa.len < 2) continue; /* should not happen */
		switch (sa.s[0]) {
		case '+':
			switch (sa.s[1]) {
			case 'a':
				set(acl->todo, TODO_BANc);
				set(todo_done, TODO_BANc);
				read_filters_addfile(&acl->bcfilter,
					&state_bcfilter, &sa.s[2]);
				break;
			case 'b':
				set(acl->todo, TODO_BAN);
				set(todo_done, TODO_BAN);
				read_filters_addfile(&acl->bfilter,
					&state_bfilter, &sa.s[2]);
				break;
			case 'B':
				set(acl->todo, TODO_BANr);
				set(todo_done, TODO_BANr);
				read_filters_addfile(&acl->brfilter,
					&state_brfilter, &sa.s[2]);
				break;
			case 'c':
				set(acl->todo, TODO_CONT);
				set(todo_done, TODO_CONT);
				read_filters_addfile(&acl->cfilter,
					&state_cfilter, &sa.s[2]);
				break;
			case 'C':
				set(acl->todo, TODO_CONTr);
				set(todo_done, TODO_CONTr);
				read_filters_addfile(&acl->crfilter,
					&state_crfilter, &sa.s[2]);
				break;
			case 'p':
				set(acl->todo, TODO_PHRASE);
				set(todo_done, TODO_PHRASE);
				read_filters_addfile(&acl->pfilter,
					&state_pfilter, &sa.s[2]);
				break;
			case 'u':
				set(acl->todo, TODO_URL);
				set(todo_done, TODO_URL);
				read_filters_addfile(&acl->ufilter,
					&state_ufilter, &sa.s[2]);
				break;
			case 'U':
				set(acl->todo, TODO_URLr);
				set(todo_done, TODO_URLr);
				read_filters_addfile(&acl->urfilter,
					&state_urfilter, &sa.s[2]);
				break;
			case 'V':
				set(acl->todo, TODO_VIRUS);
				break;
			case 'w':
				set(acl->todo, TODO_WLST);
				set(todo_done, TODO_WLST);
				read_filters_addfile(&acl->wfilter,
					&state_wfilter, &sa.s[2]);
				break;
			case 'W':
				set(acl->todo, TODO_WLSTr);
				set(todo_done, TODO_WLSTr);
				read_filters_addfile(&acl->wrfilter,
					&state_wrfilter, &sa.s[2]);
				break;
			} /* switch sa.s[1] */
			break;
		case '-':
			switch (sa.s[1]) {
			case 'a':
				unset(acl->todo, TODO_BANr);
				set(todo_done, TODO_BANr);
				state_bcfilter=STATE_NONE;
				if (!stralloc_copys(&acl->bcfilter, ""))
					die_nomem();
				break;
			case 'b':
				unset(acl->todo, TODO_BAN);
				set(todo_done, TODO_BAN);
				state_bfilter=STATE_NONE;
				if (!stralloc_copys(&acl->bfilter, ""))
					die_nomem();
				break;
			case 'B':
				unset(acl->todo, TODO_BANr);
				set(todo_done, TODO_BANr);
				state_brfilter=STATE_NONE;
				if (!stralloc_copys(&acl->brfilter, ""))
					die_nomem();
				break;
			case 'c':
				unset(acl->todo, TODO_CONT);
				set(todo_done, TODO_CONT);
				state_cfilter=STATE_NONE;
				if (!stralloc_copys(&acl->cfilter, ""))
					die_nomem();
				break;
			case 'C':
				unset(acl->todo, TODO_CONTr);
				set(todo_done, TODO_CONTr);
				state_crfilter=STATE_NONE;
				if (!stralloc_copys(&acl->crfilter, ""))
					die_nomem();
				break;
			case 'p':
				unset(acl->todo, TODO_PHRASE);
				set(todo_done, TODO_PHRASE);
				state_pfilter=STATE_NONE;
				if (!stralloc_copys(&acl->pfilter, ""))
					die_nomem();
				break;
			case 'u':
				unset(acl->todo, TODO_URL);
				set(todo_done, TODO_URL);
				state_ufilter=STATE_NONE;
				if (!stralloc_copys(&acl->ufilter, ""))
					die_nomem();
				break;
			case 'U':
				unset(acl->todo, TODO_URLr);
				set(todo_done, TODO_URLr);
				state_urfilter=STATE_NONE;
				if (!stralloc_copys(&acl->urfilter, ""))
					die_nomem();
				break;
			case 'V':
				unset(acl->todo, TODO_VIRUS);
				break;
			case 'w':
				unset(acl->todo, TODO_WLST);
				set(todo_done, TODO_WLST);
				state_wfilter=STATE_NONE;
				if (!stralloc_copys(&acl->wfilter, ""))
					die_nomem();
				break;
			case 'W':
				unset(acl->todo, TODO_WLSTr);
				set(todo_done, TODO_WLSTr);
				state_wrfilter=STATE_NONE;
				if (!stralloc_copys(&acl->wrfilter, ""))
					die_nomem();
				break;
			} /* switch sa.s[1] */
			break;
		} /* switch sa.s[0] */
	}


	/* if something is undefined, take the default */

	if (!isset(todo_done, TODO_BAN)) {
		if (isset(TODO_DEFAULT, TODO_BAN)) {
			set(acl->todo, TODO_BAN);
			state_bfilter = STATE_ALL;
		}
	}

	if (!isset(todo_done, TODO_BANc)) {
		if (isset(TODO_DEFAULT, TODO_BANc)) {
			set(acl->todo, TODO_BANc);
			state_bcfilter = STATE_ALL;
		}
	}

	if (!isset(todo_done, TODO_BANr)) {
		if (isset(TODO_DEFAULT, TODO_BANr)) {
			set(acl->todo, TODO_BANr);
			state_brfilter = STATE_ALL;
		}
	}

	if (!isset(todo_done, TODO_CONT)) {
		if (isset(TODO_DEFAULT, TODO_CONT)) {
			set(acl->todo, TODO_CONT);
			state_cfilter = STATE_ALL;
		}
	}

	if (!isset(todo_done, TODO_CONTr)) {
		if (isset(TODO_DEFAULT, TODO_CONTr)) {
			set(acl->todo, TODO_CONTr);
			state_crfilter = STATE_ALL;
		}
	}

	if (!isset(todo_done, TODO_PHRASE)) {
		if (isset(TODO_DEFAULT, TODO_PHRASE)) {
			set(acl->todo, TODO_PHRASE);
			state_pfilter = STATE_ALL;
		}
	}

	if (!isset(todo_done, TODO_URL)) {
		if (isset(TODO_DEFAULT, TODO_URL)) {
			set(acl->todo, TODO_URL);
			state_ufilter = STATE_ALL;
		}
	}

	if (!isset(todo_done, TODO_URLr)) {
		if (isset(TODO_DEFAULT, TODO_URLr)) {
			set(acl->todo, TODO_URLr);
			state_urfilter = STATE_ALL;
		}
	}

	if (!isset(todo_done, TODO_WLST)) {
		if (isset(TODO_DEFAULT, TODO_WLST)) {
			set(acl->todo, TODO_WLST);
			state_wfilter = STATE_ALL;
		}
	}

	if (!isset(todo_done, TODO_WLSTr)) {
		if (isset(TODO_DEFAULT, TODO_WLSTr)) {
			set(acl->todo, TODO_WLSTr);
			state_wrfilter = STATE_ALL;
		}
	}


	/* setup the filelists */

	if (state_bfilter == STATE_ALL) {
		list_for_each(l, &bfilter) {
			string = list_entry(l, struct squidwall_string, list);
			if(!stralloc_cat(&acl->bfilter, &string->filename))
				die_nomem();
			if(!stralloc_0(&acl->bfilter)) die_nomem();
		}
	}

	if (state_bcfilter == STATE_ALL) {
		list_for_each(l, &bcfilter) {
			regex = list_entry(l, struct squidwall_regex, list);
			if(!stralloc_cat(&acl->bcfilter, &regex->filename))
				die_nomem();
			if(!stralloc_0(&acl->bcfilter)) die_nomem();
		}
	}

	if (state_brfilter == STATE_ALL) {
		list_for_each(l, &brfilter) {
			regex = list_entry(l, struct squidwall_regex, list);
			if(!stralloc_cat(&acl->brfilter, &regex->filename))
				die_nomem();
			if(!stralloc_0(&acl->brfilter)) die_nomem();
		}
	}

	if (state_cfilter == STATE_ALL) {
		list_for_each(l, &cfilter) {
			string = list_entry(l, struct squidwall_string, list);
			if(!stralloc_cat(&acl->cfilter, &string->filename))
				die_nomem();
			if(!stralloc_0(&acl->cfilter)) die_nomem();
		}
	}

	if (state_crfilter == STATE_ALL) {
		list_for_each(l, &crfilter) {
			regex = list_entry(l, struct squidwall_regex, list);
			if(!stralloc_cat(&acl->crfilter, &regex->filename))
				die_nomem();
			if(!stralloc_0(&acl->crfilter)) die_nomem();
		}
	}

	if (state_pfilter == STATE_ALL) {
		list_for_each(l, &pfilter) {
			phrase = list_entry(l, struct squidwall_phrase, list);
			if(!stralloc_cat(&acl->pfilter, &phrase->filename))
				die_nomem();
			if(!stralloc_0(&acl->pfilter)) die_nomem();
		}
	}

	if (state_ufilter == STATE_ALL) {
		list_for_each(l, &ufilter) {
			string = list_entry(l, struct squidwall_string, list);
			if(!stralloc_cat(&acl->ufilter, &string->filename))
				die_nomem();
			if(!stralloc_0(&acl->ufilter)) die_nomem();
		}
	}

	if (state_urfilter == STATE_ALL) {
		list_for_each(l, &urfilter) {
			regex = list_entry(l, struct squidwall_regex, list);
			if(!stralloc_cat(&acl->urfilter, &regex->filename))
				die_nomem();
			if(!stralloc_0(&acl->urfilter)) die_nomem();
		}
	}

	if (state_wfilter == STATE_ALL) {
		list_for_each(l, &wfilter) {
			string = list_entry(l, struct squidwall_string, list);
			if(!stralloc_cat(&acl->wfilter, &string->filename))
				die_nomem();
			if(!stralloc_0(&acl->wfilter)) die_nomem();
		}
	}

	if (state_wrfilter == STATE_ALL) {
		list_for_each(l, &wrfilter) {
			regex = list_entry(l, struct squidwall_regex, list);
			if(!stralloc_cat(&acl->wrfilter, &regex->filename))
				die_nomem();
			if(!stralloc_0(&acl->wrfilter)) die_nomem();
		}
	}

	debug_ulong(DEBUG_READ_FILTERS, "read_filters(), acl->todo=", acl->todo);

	buffer_close(&b);
	stralloc_free(&sa);

	return;
}
